import Graph, { MultiGraph } from 'graphology';
import { Attributes } from 'graphology-types';
import { Elements, Node, Edge, XYPosition } from 'react-flow-renderer';

// Takes the querybuilder graph as an input and creates react flow elements for them.
export function createReactFlowElements(graph: Graph): Elements<Node | Edge> {
  const elements: Elements<Node | Edge> = [];

  graph.forEachNode((node: string, attributes: Attributes): void => {
    let data;
    let position = { x: attributes?.x || 0, y: attributes?.y || 0 };

    switch (attributes.type) {
      case 'entity':
        data = {
          isConnected: graph
            .neighbors(node)
            .some((nb) => graph.getNodeAttribute(nb, 'type') == 'relation'),
        };
        break;
      case 'relation':
        data = {
          isFromEntityConnected: graph
            .inNeighbors(node)
            .some((nb) => graph.getNodeAttribute(nb, 'type') == 'entity'),
          isToEntityConnected: graph
            .outNeighbors(node)
            .some((nb) => graph.getNodeAttribute(nb, 'type') == 'entity'),
        };
        break;
      case 'attribute': {
        const ERNeighbors = graph.outNeighbors(node).filter((nb) => {
          const type = graph.getNodeAttribute(nb, 'type');
          return type == 'entity' || type == 'relation';
        });
        let attributeOfA = '';
        if (ERNeighbors.length > 0)
          attributeOfA = graph.getNodeAttribute(ERNeighbors[0], 'type');
        data = {
          datatype: attributes.datatype,
          operator: attributes.operator,
          value: attributes.value,
          attributeOfA: attributeOfA,
        };
        // Get the position of the attribute, based on the connection to entity or relation
        const p = getAttributePosition(node, graph);
        if (p) position = p;
        break;
      }
    }
    // Each pill should have a name and type
    data = {
      ...data,
      name: attributes.name,
      suggestedForConnection: attributes.suggestedForConnection, // Highlights the pill, with shadow or something
    };

    const RFNode: Node = {
      id: node,
      type: attributes.type,
      position: position,
      data: data,
    };
    elements.push(RFNode);
  });

  // Add the reactflow edges
  graph.forEachEdge((edge, attributes, source, target): void => {
    // connection from attributes don't have visible connection lines
    if (attributes.type == 'attribute_connection') return;

    const RFEdge: Edge = {
      id: edge,
      source: source,
      target: target,
      type: 'connection',
      sourceHandle: attributes.sourceHandle,
      targetHandle: attributes.targetHandle,
    };
    elements.push(RFEdge);
  });

  return elements;
}

/** Gets the position of an attribute based on the connection to an entity or relation.
 * It uses the position of the parent pill and what the index is of this attribute in all
 * the connected attributes to the parent.
 */
function getAttributePosition(
  id: string,
  nodes: MultiGraph
): XYPosition | undefined {
  const nbs = nodes.filterOutNeighbors(id, (_, { type }) =>
    ['entity', 'relation'].includes(type)
  );

  if (nbs.length > 1)
    console.log(
      'WARNING: attribute connected to  more than one entity or relation'
    );
  else if (nbs.length == 1) {
    const nb = nbs[0];
    const connectedAttributes = nodes.filterInNeighbors(
      nb,
      (_, { type }) => type == 'attribute'
    );

    // An entity can have more attributes, what is the attributes index in the attributes array of that entity?
    let nthAttibute = -1;
    for (let i = 0; i < connectedAttributes.length; i++) {
      if (connectedAttributes[i] == id) {
        nthAttibute = i;
        break;
      }
    }

    const nbAttr = nodes.getNodeAttributes(nb);

    const pos = { x: nbAttr.x + 30, y: nbAttr.y + nbAttr.h };
    // ASSUMES THAT EACH ATTRIBUTE HAS THE SAME HEIGHT
    const heightOfAttributes = nodes.getNodeAttribute(id, 'h') - 1;
    pos.y += nthAttibute * heightOfAttributes;

    return pos;
  }

  // If the attribute has no (attribute_)connection, don't position it.
  return undefined;
}
